/*******************************************************************************

DAI driver by Krzysztof Strzecha and Nathan Woods


DAI technical information
==========================

CPU:
----
	8080 2MHz


Memory map:
-----------
	0000-bfff RAM
	c000-dfff ROM (non-switchable)
	e000-efff ROM (4 switchable banks)
	f000-f7ff ROM extension (optional)
	f800-f8ff SRAM (stack)
	f900-ffff I/O
		f900-faff spare
		fb00-fbff AMD9511 math chip (optional)
		fc00-fcff 8253 programmable interval timer
		fd00-fdff discrete devices
		fe00-feff 8255 PIO
		ff00-ffff timer + 5501 interrupt controller 

Interrupts:
-----------


Keyboard:
---------


Video:
-----


Sound:
------


Timings:
--------


*******************************************************************************/

#include "driver.h"
#include "cpu/i8085/i8085.h"
#include "vidhrdw/generic.h"
#include "includes/dai.h"
#include "includes/pit8253.h"
#include "machine/8255ppi.h"
#include "machine/tms5501.h"

/* I/O ports */
PORT_READ_START( dai_readport )
PORT_END

PORT_WRITE_START( dai_writeport )
PORT_END

/* memory w/r functions */
MEMORY_READ_START( dai_readmem )
	{ 0x0000, 0xbfff, MRA_BANK1 },
	{ 0xc000, 0xdfff, MRA_ROM },
	{ 0xe000, 0xefff, MRA_BANK2 },
	{ 0xf000, 0xf7ff, MRA_NOP },
	{ 0xf800, 0xf8ff, MRA_RAM },
	{ 0xfc00, 0xfcff, pit8253_0_r },
	{ 0xfd00, 0xfdff, dai_io_discrete_devices_r },
	{ 0xfe00, 0xfeff, ppi8255_0_r },
	{ 0xff00, 0xffff, tms5501_0_r },
MEMORY_END

MEMORY_WRITE_START( dai_writemem )
	{ 0x0000, 0xbfff, MWA_BANK1},
	{ 0xc000, 0xdfff, MWA_ROM},
	{ 0xe000, 0xefff, MWA_BANK2},
	{ 0xf000, 0xf7ff, MWA_NOP},
	{ 0xf800, 0xf8ff, MWA_RAM},
	{ 0xfc00, 0xfcff, pit8253_0_w },
	{ 0xfd00, 0xfdff, dai_io_discrete_devices_w },
	{ 0xfe00, 0xfeff, ppi8255_0_w },
	{ 0xff00, 0xffff, tms5501_0_w },
MEMORY_END


/* keyboard input */
INPUT_PORTS_START (dai)
	PORT_START /* [0] - 2nd PPI port A bit 0 low */
	PORT_BITX(0x01, IP_ACTIVE_LOW, IPT_KEYBOARD, "6",	KEYCODE_6,		IP_JOY_NONE )
	PORT_BITX(0x02, IP_ACTIVE_LOW, IPT_KEYBOARD, "7",	KEYCODE_7,		IP_JOY_NONE )
	PORT_BITX(0x04, IP_ACTIVE_LOW, IPT_KEYBOARD, "8",	KEYCODE_8,		IP_JOY_NONE )
	PORT_BITX(0x08, IP_ACTIVE_LOW, IPT_KEYBOARD, "Ready",	KEYCODE_INSERT,		IP_JOY_NONE )
	PORT_BITX(0x10, IP_ACTIVE_LOW, IPT_KEYBOARD, "Tab",	KEYCODE_TAB,		IP_JOY_NONE )
	PORT_BITX(0x20, IP_ACTIVE_LOW, IPT_KEYBOARD, "=",	KEYCODE_EQUALS,		IP_JOY_NONE )
	PORT_BITX(0x40, IP_ACTIVE_LOW, IPT_KEYBOARD, "0",	KEYCODE_0,		IP_JOY_NONE )
	PORT_BITX(0x80, IP_ACTIVE_LOW, IPT_KEYBOARD, "9",	KEYCODE_9,		IP_JOY_NONE )

	PORT_START /* [1] - 2nd PPI port A bit 1 low */
	PORT_BITX(0x01, IP_ACTIVE_LOW, IPT_KEYBOARD, "G",	KEYCODE_G,		IP_JOY_NONE )
	PORT_BITX(0x02, IP_ACTIVE_LOW, IPT_KEYBOARD, "[",	KEYCODE_OPENBRACE,	IP_JOY_NONE )
	PORT_BITX(0x04, IP_ACTIVE_LOW, IPT_KEYBOARD, "]",	KEYCODE_CLOSEBRACE,	IP_JOY_NONE )
	PORT_BITX(0x08, IP_ACTIVE_LOW, IPT_KEYBOARD, "Enter",	KEYCODE_ENTER,		IP_JOY_NONE )
	PORT_BITX(0x10, IP_ACTIVE_LOW, IPT_KEYBOARD, "Run",	KEYCODE_DEL,		IP_JOY_NONE )
	PORT_BITX(0x20, IP_ACTIVE_LOW, IPT_KEYBOARD, "*",	KEYCODE_QUOTE,		IP_JOY_NONE )
	PORT_BITX(0x40, IP_ACTIVE_LOW, IPT_KEYBOARD, "H",	KEYCODE_H,		IP_JOY_NONE )
	PORT_BITX(0x80, IP_ACTIVE_LOW, IPT_KEYBOARD, "Z",	KEYCODE_Z,		IP_JOY_NONE )

	PORT_START /* [2] - 2nd PPI port A bit 2 low */
	PORT_BITX(0x01, IP_ACTIVE_LOW, IPT_KEYBOARD, "R",	KEYCODE_R,		IP_JOY_NONE )
	PORT_BITX(0x02, IP_ACTIVE_LOW, IPT_KEYBOARD, "O",	KEYCODE_O,		IP_JOY_NONE )
	PORT_BITX(0x04, IP_ACTIVE_LOW, IPT_KEYBOARD, "L",	KEYCODE_L,		IP_JOY_NONE )
	PORT_BITX(0x08, IP_ACTIVE_LOW, IPT_KEYBOARD, "Del",	KEYCODE_BACKSPACE,	IP_JOY_NONE )
	PORT_BITX(0x10, IP_ACTIVE_LOW, IPT_KEYBOARD, ".",	KEYCODE_STOP,		IP_JOY_NONE )
	PORT_BITX(0x20, IP_ACTIVE_LOW, IPT_KEYBOARD, "\\",	KEYCODE_BACKSLASH,	IP_JOY_NONE )
	PORT_BITX(0x40, IP_ACTIVE_LOW, IPT_KEYBOARD, "V",	KEYCODE_V,		IP_JOY_NONE )
	PORT_BITX(0x80, IP_ACTIVE_LOW, IPT_KEYBOARD, "D",	KEYCODE_D,		IP_JOY_NONE )

	PORT_START /* [3] - 2nd PPI port A bit 3 low */
	PORT_BITX(0x01, IP_ACTIVE_LOW, IPT_KEYBOARD, "Space",	KEYCODE_SPACE,		IP_JOY_NONE )
	PORT_BITX(0x02, IP_ACTIVE_LOW, IPT_KEYBOARD, "B",	KEYCODE_B,		IP_JOY_NONE )
	PORT_BITX(0x04, IP_ACTIVE_LOW, IPT_KEYBOARD, "@",	KEYCODE_ESC,		IP_JOY_NONE )
	PORT_BITX(0x08, IP_ACTIVE_LOW, IPT_KEYBOARD, "Alt",	KEYCODE_RSHIFT,		IP_JOY_NONE )
	PORT_BITX(0x10, IP_ACTIVE_LOW, IPT_KEYBOARD, "_",	KEYCODE_MINUS,		IP_JOY_NONE )
	PORT_BITX(0x20, IP_ACTIVE_LOW, IPT_KEYBOARD, "Latin",	KEYCODE_RCONTROL,	IP_JOY_NONE )
	PORT_BITX(0x40, IP_ACTIVE_LOW, IPT_KEYBOARD, "/",	KEYCODE_SLASH,		IP_JOY_NONE )
	PORT_BITX(0x80, IP_ACTIVE_LOW, IPT_KEYBOARD, ",",	KEYCODE_COMMA,		IP_JOY_NONE )

	PORT_START /* [4] - 2nd PPI port A bit 4 low */
	PORT_BITX(0x01, IP_ACTIVE_LOW, IPT_KEYBOARD, "Cls",	KEYCODE_END,		IP_JOY_NONE )
	PORT_BITX(0x02, IP_ACTIVE_LOW, IPT_KEYBOARD, "(G)",	KEYCODE_F1,		IP_JOY_NONE )
	PORT_BITX(0x04, IP_ACTIVE_LOW, IPT_KEYBOARD, "(B)",	KEYCODE_F2,		IP_JOY_NONE )
	PORT_BITX(0x08, IP_ACTIVE_LOW, IPT_KEYBOARD, "5",	KEYCODE_5,		IP_JOY_NONE )
	PORT_BITX(0x10, IP_ACTIVE_LOW, IPT_KEYBOARD, "4",	KEYCODE_4,		IP_JOY_NONE )
	PORT_BITX(0x20, IP_ACTIVE_LOW, IPT_KEYBOARD, "3",	KEYCODE_3,		IP_JOY_NONE )
	PORT_BITX(0x40, IP_ACTIVE_LOW, IPT_KEYBOARD, "2",	KEYCODE_2,		IP_JOY_NONE )
	PORT_BITX(0x80, IP_ACTIVE_LOW, IPT_KEYBOARD, "1",	KEYCODE_1,		IP_JOY_NONE )

	PORT_START /* [5] - 2nd PPI port A bit 5 low */
	PORT_BIT (0x01, IP_ACTIVE_LOW, IPT_UNUSED)
	PORT_BIT (0x02, IP_ACTIVE_LOW, IPT_UNUSED)
	PORT_BITX(0x04, IP_ACTIVE_LOW, IPT_KEYBOARD, "J",	KEYCODE_J,		IP_JOY_NONE )
	PORT_BITX(0x08, IP_ACTIVE_LOW, IPT_KEYBOARD, "N",	KEYCODE_N,		IP_JOY_NONE )
	PORT_BITX(0x10, IP_ACTIVE_LOW, IPT_KEYBOARD, "E",	KEYCODE_E,		IP_JOY_NONE )
	PORT_BITX(0x20, IP_ACTIVE_LOW, IPT_KEYBOARD, "K",	KEYCODE_K,		IP_JOY_NONE )
	PORT_BITX(0x40, IP_ACTIVE_LOW, IPT_KEYBOARD, "U",	KEYCODE_U,		IP_JOY_NONE )
	PORT_BITX(0x80, IP_ACTIVE_LOW, IPT_KEYBOARD, "C",	KEYCODE_C,		IP_JOY_NONE )

	PORT_START /* [6] - 2nd PPI port A bit 6 low */
	PORT_BITX(0x01, IP_ACTIVE_LOW, IPT_KEYBOARD, ";",	KEYCODE_COLON,		IP_JOY_NONE )
	PORT_BITX(0x02, IP_ACTIVE_LOW, IPT_KEYBOARD, "Russian",	KEYCODE_LCONTROL,	IP_JOY_NONE )
	PORT_BITX(0x04, IP_ACTIVE_LOW, IPT_KEYBOARD, "Key",	KEYCODE_CAPSLOCK,	IP_JOY_NONE )
	PORT_BITX(0x08, IP_ACTIVE_LOW, IPT_KEYBOARD, "P",	KEYCODE_P,		IP_JOY_NONE )
	PORT_BITX(0x10, IP_ACTIVE_LOW, IPT_KEYBOARD, "A",	KEYCODE_A,		IP_JOY_NONE )
	PORT_BITX(0x20, IP_ACTIVE_LOW, IPT_KEYBOARD, "W",	KEYCODE_W,		IP_JOY_NONE )
	PORT_BITX(0x40, IP_ACTIVE_LOW, IPT_KEYBOARD, "Y",	KEYCODE_Y,		IP_JOY_NONE )
	PORT_BITX(0x80, IP_ACTIVE_LOW, IPT_KEYBOARD, "F",	KEYCODE_F,		IP_JOY_NONE )

	PORT_START /* [7] - 2nd PPI port A bit 7 low */
	PORT_BITX(0x01, IP_ACTIVE_LOW, IPT_KEYBOARD, "Shift",	KEYCODE_LSHIFT,		IP_JOY_NONE )
	PORT_BITX(0x02, IP_ACTIVE_LOW, IPT_KEYBOARD, "Q",	KEYCODE_Q,		IP_JOY_NONE )
	PORT_BITX(0x04, IP_ACTIVE_LOW, IPT_KEYBOARD, "^",	KEYCODE_TILDE,		IP_JOY_NONE )
	PORT_BITX(0x08, IP_ACTIVE_LOW, IPT_KEYBOARD, "X",	KEYCODE_X,		IP_JOY_NONE )
	PORT_BITX(0x10, IP_ACTIVE_LOW, IPT_KEYBOARD, "T",	KEYCODE_T,		IP_JOY_NONE )
	PORT_BITX(0x20, IP_ACTIVE_LOW, IPT_KEYBOARD, "I",	KEYCODE_I,		IP_JOY_NONE )
	PORT_BITX(0x40, IP_ACTIVE_LOW, IPT_KEYBOARD, "M",	KEYCODE_M,		IP_JOY_NONE )
	PORT_BITX(0x80, IP_ACTIVE_LOW, IPT_KEYBOARD, "S",	KEYCODE_S,		IP_JOY_NONE )

	PORT_START /* [8] - 2nd PPI port C bit 0 low */
	PORT_BITX(0x10, IP_ACTIVE_LOW, IPT_KEYBOARD, "PrnLck",	KEYCODE_F6,		IP_JOY_NONE )
	PORT_BITX(0x20, IP_ACTIVE_LOW, IPT_KEYBOARD, "ScrLck",	KEYCODE_F5,		IP_JOY_NONE )
	PORT_BITX(0x40, IP_ACTIVE_LOW, IPT_KEYBOARD, "Sound",	KEYCODE_F4,		IP_JOY_NONE )
	PORT_BITX(0x80, IP_ACTIVE_LOW, IPT_KEYBOARD, "(R)",	KEYCODE_F3,		IP_JOY_NONE )

	PORT_START /* [9] - 2nd PPI port C bit 1 low */
	PORT_BITX(0x10, IP_ACTIVE_LOW, IPT_KEYBOARD, "ScrPrn",	KEYCODE_F7,		IP_JOY_NONE )
	PORT_BITX(0x20, IP_ACTIVE_LOW, IPT_KEYBOARD, "F0",	KEYCODE_F8,		IP_JOY_NONE )
	PORT_BITX(0x40, IP_ACTIVE_LOW, IPT_KEYBOARD, "F1",	KEYCODE_F9,		IP_JOY_NONE )
	PORT_BITX(0x80, IP_ACTIVE_LOW, IPT_KEYBOARD, "F2",	KEYCODE_F10,		IP_JOY_NONE )

	PORT_START /* [10] - 2nd PPI port C bit 2 low */
	PORT_BITX(0x10, IP_ACTIVE_LOW, IPT_KEYBOARD, "Home",	KEYCODE_HOME,		IP_JOY_NONE )
	PORT_BITX(0x20, IP_ACTIVE_LOW, IPT_KEYBOARD, "F5",	KEYCODE_SCRLOCK,	IP_JOY_NONE )
	PORT_BITX(0x40, IP_ACTIVE_LOW, IPT_KEYBOARD, "F4",	KEYCODE_F12,		IP_JOY_NONE )
	PORT_BITX(0x80, IP_ACTIVE_LOW, IPT_KEYBOARD, "F3",	KEYCODE_F11,		IP_JOY_NONE )

	PORT_START /* [12] - 2nd PPI port C bit 3 low */
	PORT_BITX(0x10, IP_ACTIVE_LOW, IPT_KEYBOARD, "Right",	KEYCODE_RIGHT,		IP_JOY_NONE )
	PORT_BITX(0x20, IP_ACTIVE_LOW, IPT_KEYBOARD, "Up",	KEYCODE_UP,		IP_JOY_NONE )
	PORT_BITX(0x40, IP_ACTIVE_LOW, IPT_KEYBOARD, "Left",	KEYCODE_LEFT,		IP_JOY_NONE )
	PORT_BITX(0x80, IP_ACTIVE_LOW, IPT_KEYBOARD, "Down",	KEYCODE_DOWN,		IP_JOY_NONE )

	PORT_START /* [13] - CPU */
	PORT_BITX(0x01, IP_ACTIVE_HIGH, IPT_KEYBOARD, "Reset",	KEYCODE_PGUP,		IP_JOY_NONE )

	PORT_START /* [14] - CPU */
	PORT_BIT(0x04, IP_ACTIVE_HIGH, IPT_VBLANK)
	PORT_BIT(0x10, IP_ACTIVE_HIGH, IPT_BUTTON1 | IPF_PLAYER1)
	PORT_BIT(0x20, IP_ACTIVE_HIGH, IPT_BUTTON1 | IPF_PLAYER2)
	PORT_BIT(0xcb, IP_ACTIVE_HIGH, IPT_UNUSED)
INPUT_PORTS_END


/* machine definition */
static MACHINE_DRIVER_START( dai )
	/* basic machine hardware */
	MDRV_CPU_ADD(8080, 2000000)
	MDRV_CPU_MEMORY(dai_readmem, dai_writemem)
	MDRV_CPU_PORTS(dai_readport, dai_writeport)
	MDRV_FRAMES_PER_SECOND(50)
	MDRV_VBLANK_DURATION(0)
	MDRV_INTERLEAVE(1)

	MDRV_MACHINE_INIT( dai )

    /* video hardware */
	MDRV_VIDEO_ATTRIBUTES(VIDEO_TYPE_RASTER)
	MDRV_SCREEN_SIZE(528, 260)
	MDRV_VISIBLE_AREA(0, 528-1, 0, 260-1)
	MDRV_PALETTE_LENGTH(sizeof (dai_palette) / 3)
	MDRV_COLORTABLE_LENGTH(sizeof (dai_colortable))
	MDRV_PALETTE_INIT( dai )

	MDRV_VIDEO_START( dai )
	MDRV_VIDEO_UPDATE( dai )
MACHINE_DRIVER_END

#define io_dai		io_NULL

ROM_START(dai)
	ROM_REGION(0x14000,REGION_CPU1,0)
	ROM_LOAD("dai.bin", 0xc000, 0x2000, CRC(ca71a7d5))
	ROM_LOAD("dai00.bin", 0x10000, 0x1000, CRC(fa7d39ac))
	ROM_LOAD("dai01.bin", 0x11000, 0x1000, CRC(cb5809f2))
	ROM_LOAD("dai02.bin", 0x12000, 0x1000, CRC(03f72d4a))
	ROM_LOAD("dai03.bin", 0x13000, 0x1000, CRC(c475c96f))
	ROM_REGION(0x2000, REGION_GFX1,0)
	ROM_LOAD ("nch.bin", 0x0000, 0x2000, CRC(0042d40d))
ROM_END


SYSTEM_CONFIG_START(dai)
	CONFIG_RAM_DEFAULT(48 * 1024)
SYSTEM_CONFIG_END


/*     YEAR  NAME	PARENT  COMPAT	MACHINE	INPUT	INIT	CONFIG	COMPANY	FULLNAME */
COMPX( 1989, dai,	0,      0,		dai,	dai,	0,		dai,	"",		"DAI", GAME_NOT_WORKING )
